/*
 * Copyright (c) 2002 by Tsuyoshi Iguchi (zinnia@risky-safety.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id$
 */

#include <stdlib.h>
#include <SDL.h>
#include <SDL_thread.h>
#include "SDLSaveMovie.h"
#include "SDLSaveMovie_int.h"

#ifdef USE_JPG
#include "SDLSaveJPG.h"
#endif

#ifdef USE_OPENGL
#include "GL/gl.h"
#endif

callback_vector* vector[] = {
#ifdef USE_MPEG
	&SDLSaveMovie_MPEG_vector,
#endif
#ifdef USE_SWF
	&SDLSaveMovie_SWF_vector,
#endif
#ifdef USE_MNG
	&SDLSaveMovie_MNG_vector,
#endif
	NULL
};

/* error context */
static char* errormessage = "";

static callback_vector* query_driver(char* name)
{
	callback_vector** p;
	for(p = vector; *p != NULL; p++){
		if(strcmp((*p)->identifier, name) == 0) return *p;
	}
	return NULL;
}

static int driver_getprivatesize()
{
	int result = 0;
	int tmp;
	callback_vector** p;
	for(p = vector; *p != NULL; p++){
		tmp = (*p)->callback_getprivatesize();
		if(result < tmp) result = tmp;
	}
	return result;
}
    
static int implicit_initialize_driver(SDLSaveMovie* param)
{
	callback_vector* p;
	char* env;
	SDLSaveMovie_context* ctx = param->internal;

	if(ctx->callback != NULL) return 1;

	if(vector[0] == NULL){
		SDLSaveMovie_SetError("no driver available.");
		return -1;
	}
	if(param->drivername != NULL){
		p = query_driver(param->drivername);
		if(p != NULL){
			ctx->callback = p;
			return 0;
		}
		if(param->driver_fallback == 0) return 1;
	}
	env = getenv("SDLSAVEMOVIE_DRIVER");
	if(env != NULL){
		p = query_driver(env);
		if(p != NULL){
			ctx->callback = p;
			return 0;
		}
		if(param->driver_fallback == 0) return 1;
	}
	ctx->callback = vector[0];
	return 0;
}

SDLSaveMovie* SDLSaveMovie_Init()
{
	int privatesize;
	SDLSaveMovie* p;
	p = calloc(sizeof(SDLSaveMovie), 0);
	if(p == NULL) return NULL;

	p->interval = 100; // 10fps
	p->quality = 50;
	p->threadingmodel = SDLSAVEMOVIE_MODEL_THREAD;
	p->useaudio = 0;
	p->savepath = NULL;
	p->drivername = NULL;
	p->driver_fallback = 1; // true
	p->rwops_needsfree = 0; // false
	privatesize = driver_getprivatesize();
    
	p->internal = calloc(sizeof(SDLSaveMovie_context), 0);
	if(p->internal == NULL){
		free(p);
		return NULL;
	}

	if(privatesize > 0){
		p->private = calloc(privatesize, 0);
		if(p->private == NULL){
			free(p->internal);
			free(p);
			return NULL;
		}
	} else {
		p->private = NULL;
	}
	return p;
}

void SDLSaveMovie_Terminate(SDLSaveMovie* param)
{
	if(param->private != NULL) free(param->private);
	if(param->internal != NULL) free(param->internal);
	free(param);
}

void SDLSaveMovie_Update(SDLSaveMovie*  param )
{
	struct SDLSaveMovie_context* ctx = param->internal;

	if(ctx->srctype == SDLSAVEMOVIE_SRC_OPENGL){
#ifdef USE_OPENGL
		glReadPixels(ctx->rect.x, ctx->rect.y, ctx->rect.w, ctx->rect.h, GL_RGB, GL_UNSIGNED_BYTE, ctx->source.glpixels);
      
		switch(ctx->requesttype){
		case SDLSAVEMOVIE_REQUEST_RAW:
			ctx->size = ctx->rect.w * ctx->rect.h * 3;
			ctx->ptr = ctx->source.glpixels;
			break;
#ifdef USE_JPG
		case SDLSAVEMOVIE_REQUEST_JPG:
			SDL_RWseek(ctx->rwmem, 0, SEEK_SET);
			SDLSaveJPG_GL_RW(ctx->source.glpixels, ctx->rect.h, ctx->rect.w, ctx->rwmem, param->quality, 0);
			ctx->size = SDL_RWtell(ctx->rwmem);
			ctx->ptr = ctx->jpgtmp;
			break;
#endif
		}

#else // USE_OPENGL
		fprintf(stderr, "SDLSaveMovie: *fatal* you can't grab OpenGL pixbuf.\n");
#endif
	} else {
		// SDLSAVEMOVIE_SRC_SDLSURFACE
		switch(ctx->requesttype){
		case SDLSAVEMOVIE_REQUEST_RAW:
			ctx->size = ctx->pitch * ctx->rect.h;
			ctx->ptr = ctx->source.surface;
			break;
#ifdef USE_JPG
		case SDLSAVEMOVIE_REQUEST_JPG:
			SDL_RWseek(ctx->rwmem, 0, SEEK_SET);
			SDLSaveJPG_RW(ctx->source.surface, ctx->rwmem, param->quality, 0);
			ctx->size = SDL_RWtell(ctx->rwmem);
			ctx->ptr = ctx->jpgtmp;
			break;
#endif
		}

	}
	ctx->callback->callback_main(param);
  
}

static int ThreadFunction(void* data)
{
	Uint32 begintick, nowtick;
	int ticktowait;
	Uint32 tickexceeds_max = 0;
	SDLSaveMovie* param = (SDLSaveMovie*) data;

	while(param->internal->running == 1){
		begintick = SDL_GetTicks();

		SDLSaveMovie_Update(param);

		nowtick = SDL_GetTicks();
		ticktowait = param->interval - (nowtick - begintick);
		if(ticktowait > 0) {
			SDL_Delay(ticktowait);
		} else {
			if(tickexceeds_max < -ticktowait){
				tickexceeds_max = -ticktowait;
				fprintf(stderr, "SDLSaveMovie: thread: process time exceeds %d ticks.\n", tickexceeds_max);
			}
		}
	}
	return 0;
}

static int SDLSaveMovie_Start_common(SDL_RWops* rwops, SDLSaveMovie* param)
{
	SDLSaveMovie_context* ctx = param->internal;

	if(implicit_initialize_driver(param) < 0) return -1;
	switch(param->threadingmodel){
	case SDLSAVEMOVIE_MODEL_THREAD:
	case SDLSAVEMOVIE_MODEL_PUMP:
		break;
	default:
		SDLSaveMovie_SetError("invalid model");
		return -1;
	}

	ctx->rwops = rwops;
	if(param->quality > 100) param->quality = 100;
	else if(param->quality < 0) param->quality = 0;

	if(ctx->callback->callback_init(param) != 0){
		SDLSaveMovie_SetError("initalizing callback failed.");
		return -1;
	}

#ifdef USE_JPG
	if(ctx->requesttype == SDLSAVEMOVIE_REQUEST_JPG){
		int jpgsize = param->internal->pitch * param->internal->rect.h;
		ctx->jpgtmp = (Uint8*)malloc(jpgsize);
		if(ctx->jpgtmp == NULL) return -1;
		ctx->rwmem = SDL_RWFromMem(ctx->jpgtmp, jpgsize);
		if(ctx->rwmem == NULL) return -1;
	}
#endif

	ctx->running = 1;
	if(param->threadingmodel == SDLSAVEMOVIE_MODEL_THREAD){
		ctx->mythread = SDL_CreateThread(ThreadFunction, (void*)param);
		if(ctx->mythread == NULL) {
			SDLSaveMovie_SetError("initializing thread failed.");
			return -1;
		}
	}
	return 0;
}

static int init_ctx(SDLSaveMovie* param){
	SDLSaveMovie_context* ctx = param->internal;

	if(ctx->running != 0){
		SDLSaveMovie_SetError("already running.");
		return -1;
	}
	return 0;
}

int SDLSaveMovie_Start_SDL_RW(SDLSaveMovie* param, SDL_Surface* src, SDL_RWops* rwops)
{
	SDLSaveMovie_context* ctx = param->internal;

	if(init_ctx(param) != 0) return -1;

	ctx->srctype = SDLSAVEMOVIE_SRC_SDLSURFACE;
	ctx->source.surface = src;
	ctx->rect.w = ctx->source.surface->w;
	ctx->rect.h = ctx->source.surface->h;
	ctx->rect.x = ctx->rect.y = 0;

	ctx->pitch = ctx->source.surface->pitch;

	SDL_GetRGB(src->format->colorkey, src->format, &(ctx->r), &(ctx->g), &(ctx->b));
	return SDLSaveMovie_Start_common(rwops, param);
}

static char* compose_filename(char* base, SDLSaveMovie* param)
{
	char* ext;
	int l;

	if(implicit_initialize_driver(param) < 0) return NULL;

	ext = param->internal->callback->callback_getextension();
	l = strlen(base) + strlen(ext) + 2;

	param->savepath = malloc(l + 1); // XXX should be freed.
	if(param->savepath == NULL) return NULL;
	snprintf(param->savepath, l, "%s.%s", base, ext);
	return param->savepath;
}

int SDLSaveMovie_Start_SDL(SDLSaveMovie* param, SDL_Surface* src, char* filename)
{
	if(compose_filename(filename, param) == NULL) return -1;
	param->rwops_needsfree = 1;

	return SDLSaveMovie_Start_SDL_RW(param, src, SDL_RWFromFile(param->savepath, "wb"));
}

int SDLSaveMovie_Start_GL_RW(SDLSaveMovie* param, SDL_Rect* rect, Uint8 r, Uint8 g, Uint8 b, SDL_RWops* rwops)
{
	SDLSaveMovie_context* ctx = param->internal;
#ifdef USE_OPENGL
	if(init_ctx(param) != 0) return -1;
	ctx->srctype = SDLSAVEMOVIE_SRC_OPENGL;
	ctx->rect.w = rect->w;
	ctx->rect.h = rect->h;
	ctx->rect.x = rect->x;
	ctx->rect.y = rect->y;

	ctx->pitch = ctx->rect.w * 3;

	ctx->r = r; ctx->g = g; ctx->b = b;
	ctx->source.glpixels = malloc(ctx->rect.w * ctx->rect.h * 4); // XXX 32bpp
	if(ctx->source.glpixels == NULL) return -1;

	return SDLSaveMovie_Start_common(rwops, param);
#else
	SDLSaveMovie_SetError("OpenGL not supported.");
	return -1;
#endif
}

int SDLSaveMovie_Start_GL(SDLSaveMovie* param, SDL_Rect* rect, Uint8 r, Uint8 g, Uint8 b, char* filename)
{
	if(compose_filename(filename, param) == NULL) return -1;
	param->rwops_needsfree = 1;
	return SDLSaveMovie_Start_GL_RW(param, rect, r, g, b, SDL_RWFromFile(param->savepath, "wb"));
}

  
void SDLSaveMovie_Stop(SDLSaveMovie* param)
{
	SDLSaveMovie_context* ctx = param->internal;

	if(param->threadingmodel == SDLSAVEMOVIE_MODEL_THREAD){
		ctx->running = 0;
		SDL_WaitThread(ctx->mythread, NULL);
	}

	ctx->callback->callback_terminate(param);

#ifdef USE_OPENGL
	if(ctx->srctype == SDLSAVEMOVIE_SRC_OPENGL){
		free(ctx->source.glpixels);
	}
#endif
#ifdef USE_JPG
	if(ctx->requesttype == SDLSAVEMOVIE_REQUEST_JPG){
		SDL_FreeRW(ctx->rwmem);
		free(ctx->jpgtmp);
	}
#endif
	if(param->rwops_needsfree == 1){
		SDL_FreeRW(param->internal->rwops);
	}
	if(param->savepath != NULL) free(param->savepath);

	return;
}


char* SDLSaveMovie_GetError()
{
	return errormessage;
}

void SDLSaveMovie_SetError(char* message)
{
	errormessage = message;
}
